#include "StaticString.h"
#include "ConceptTypes.h"
#include "SHManager.h"

POOLED_IMPLEMENTATION(StaticString)

#ifdef USE_DEPRECATED
    #include <string.h>
    #include <stdlib.h>
    #include <stdio.h>
#else
    #include <string.h>
    #include <stdio.h>
    #include <stdlib.h>
#endif


#define breakeven_point  12
#define fast_memcpy(d,s,n) \
  { register size_t nn = (size_t)(n); \
    if (nn >= breakeven_point) memcpy((d), (s), nn); \
    else if (nn > 0) { /* proc call overhead is worth only for large strings*/\
      register char *dd; register const char *ss; \
      for (ss=(s), dd=(d); nn>0; nn--) *dd++ = *ss++; } }

#define MEMCPY      fast_memcpy

StaticString::StaticString() {
    DataOffset=0;
    _LENGTH=0;
}

StaticString::StaticString(char *value) {
    DataOffset=0;
    _LENGTH=0;
    operator=(value);
}

StaticString::StaticString(const StaticString &S) {
    DataOffset=0;
    _LENGTH=0;
    if (S._LENGTH) {
        char *Data=0;
        if (S.DataOffset)
            Data=((char *)(void *)&S)+S.DataOffset;
        this->LoadBuffer(Data,S._LENGTH);
    }
}


void StaticString::operator=(char *value) {
    char *Data=0;
    if (DataOffset)
        Data=((char *)(void *)this)+DataOffset;

    size_t len;
    if (Data) {
        free(Data);
        Data=0;
        DataOffset=0;
        _LENGTH=0;
    }
    if (value) {
        len=strlen(value);
        _LENGTH=len;
        if (len) {
            Data=(char *)realloc(Data,len+1);
            MEMCPY(Data,value,len+1)

            DataOffset=(long)Data-(long)this;
        }
    }
}

void StaticString::operator=(AnsiString S) {
    this->LoadBuffer(S.c_str(), S.Length());
}

void StaticString::operator=(StaticString S) {
    size_t len;

    char *Data=0;
    if (DataOffset)
        Data=((char *)(void *)this)+DataOffset;

    if (Data)
        Data[0]=0;
    _LENGTH=0;

    char *other_data=S.c_str();
    len=S.Length();
    if (len) {
        _LENGTH=len;
        free(Data);
        Data=(char *)malloc(len+1);
        MEMCPY(Data,other_data,len+1);

        DataOffset=(long)Data-(long)this;
    }
}

StaticString::~StaticString(void)
{
    char *Data=0;
    if (DataOffset)
        Data=((char *)(void *)this)+DataOffset;

    free(Data);
}

StaticString::operator char *() {
    return c_str();
}

char *StaticString::c_str() {
    char *Data=0;
    if (!_LENGTH)
        return "";

    if (DataOffset)
        Data=((char *)(void *)this)+DataOffset;

    if (!Data)
        return "";
    return Data;
}

long StaticString::ToInt() {
    char *Data=0;
    if (DataOffset)
        Data=((char *)(void *)this)+DataOffset;

    if (Data)
        return atol(Data);

    return 0;
}

double StaticString::ToFloat() {
    char *Data=0;
    if (DataOffset)
        Data=((char *)(void *)this)+DataOffset;

    if (Data)
        return atof(Data);
    return 0;
}

long StaticString::Length() {
    char *Data=0;
    if (DataOffset)
        Data=((char *)(void *)this)+DataOffset;

    if (!Data)
        return 0;
    return _LENGTH;
}

void StaticString::LoadBuffer(char *buffer, int size) {
    char *Data=0;
    if (DataOffset)
        Data=((char *)(void *)this)+DataOffset;

    free(Data);
    Data=0;
    _LENGTH=size;

    if (size) {
        Data=(char *)malloc(size+1);
        MEMCPY(Data,buffer,size);
        Data[size]=0;

        DataOffset=(long)Data-(long)this;
    }
}

int StaticString::Serialize(FILE *out, int type) {
    unsigned char uClen=(unsigned char)_LENGTH;
    unsigned short uSlen=(unsigned short)_LENGTH;
    char *Data=0;
    if (DataOffset)
        Data=((char *)(void *)this)+DataOffset;

    switch (type) {
        case SERIALIZE_8BIT_LENGTH :
            fwrite(&uClen,sizeof(uClen),1,out);
            break;
        case SERIALIZE_16BIT_LENGTH :
            fwrite(&uSlen,sizeof(uSlen),1,out);
            break;
        default :
            fwrite(&_LENGTH,sizeof(_LENGTH),1,out);
            break;
    }
    fwrite(Data,_LENGTH,1,out);
    return 1;
}

int StaticString::ComputeSharedSize(FILE *in, int type) {
    int len;
    unsigned char uClen=0;
    unsigned short uSlen=0;

    switch (type) {
        case SERIALIZE_8BIT_LENGTH :
            FREAD_FAIL(&uClen,sizeof(uClen),1,in);
            len=uClen;
            break;
        case SERIALIZE_16BIT_LENGTH :
            FREAD_FAIL(&uSlen,sizeof(uSlen),1,in);
            len=uSlen;
            break;
        default:
            FREAD_FAIL(&len,sizeof(len),1,in);
            break;
    }
    SKIP(len, in);
    return len+1;
}

int StaticString::Unserialize(FILE *in, int type, char use_pool) {
    int len;
    unsigned char uClen=(unsigned char)_LENGTH;
    unsigned short uSlen=(unsigned short)_LENGTH;
    char *Data=0;
    if (DataOffset)
        Data=((char *)(void *)this)+DataOffset;

    switch (type) {
        case SERIALIZE_8BIT_LENGTH :
            if (!fread(&uClen,sizeof(uClen),1,in))
                return -1;
            len=uClen;
            break;
        case SERIALIZE_16BIT_LENGTH :
            if (!fread(&uSlen,sizeof(uSlen),1,in))
                return -1;
            len=uSlen;
            break;
        default :
            if (!fread(&len,sizeof(len),1,in))
                return -1;
            break;
    }

    free(Data);
    Data=0;
    if (len) {
        if (use_pool==1) {
            _LENGTH=len;
            Data=(char *)SHAlloc(len+1);
            fread(Data,len,1,in);
            Data[len]=0;
            DataOffset=(long)Data-(long)this;
        } else {
            _LENGTH=len;
            Data=(char *)malloc(len+1);
            fread(Data,len,1,in);
            Data[len]=0;
            // moving the top pointer
            if (use_pool==-1)
                SHAlloc(len+1);
            DataOffset=(long)Data-(long)this;
        }
    } else
        _LENGTH=0;
    return 1;
}
